"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var language_1 = require("graphql/language");
var GraphQLError_1 = require("graphql/error/GraphQLError");
var getOperationAST_1 = require("graphql/utilities/getOperationAST");
var values_1 = require("graphql/execution/values");
var get = require("lodash/get.js");
var set = require("lodash/set.js");
var each = require("lodash/each.js");
var keyBy = require("lodash/keyBy.js");
var isEqual = require("lodash/isEqual.js");
var transformations_1 = require("./transformations");
var schema_1 = require("./schema");
function graphqlLodash(query, operationName) {
    var pathToArgs = {};
    var reshape = {};
    var queryAST = typeof query === 'string' ? language_1.parse(query) : query;
    traverseOperationFields(queryAST, operationName, function (node, resultPath, alias) {
        var args = getLodashDirectiveArgs(node, schema_1.lodashAST());
        if (args === null)
            return;
        var argsSetPath = resultPath.concat(['@_']);
        var previousArgsValue = get(pathToArgs, argsSetPath, null);
        if (previousArgsValue !== null && !isEqual(previousArgsValue, args))
            throw Error("Different \"@_\" args for the \"" + resultPath.join('.') + "\" path");
        set(pathToArgs, argsSetPath, args);
        set(reshape, resultPath.concat(['_@']), alias[resultPath.length - 1]);
    });
    return {
        apply: !!Object.keys(pathToArgs).length,
        query: stripQuery(queryAST),
        transform: function (data) { return applyLodashDirective(pathToArgs, data); },
        queryShape: pathToArgs,
        reshape: reshape,
    };
}
exports.graphqlLodash = graphqlLodash;
function traverseOperationFields(queryAST, operationName, cb) {
    var fragments = {};
    var operationAST = getOperationAST_1.getOperationAST(queryAST, operationName);
    if (!operationAST) {
        throw new GraphQLError_1.GraphQLError('Must provide operation name if query contains multiple operations.');
    }
    queryAST.definitions.forEach(function (definition) {
        if (definition.kind === language_1.Kind.FRAGMENT_DEFINITION)
            fragments[definition.name.value] = definition;
    });
    var resultPath = [];
    var alias = [];
    cb(operationAST, resultPath, alias);
    traverse(operationAST);
    function traverse(root) {
        language_1.visit(root, {
            enter: function (node) {
                if (node.kind === language_1.Kind.FIELD) {
                    resultPath.push((node.alias || node.name).value);
                    alias.push((_a = {}, _a[(node.alias || node.name).value] = node.name.value, _a));
                }
                if (node.kind === language_1.Kind.FRAGMENT_SPREAD) {
                    var fragmentName = node.name.value;
                    var fragment = fragments[fragmentName];
                    if (!fragment)
                        throw Error("Unknown fragment: " + fragmentName);
                    traverse(fragment);
                }
                var _a;
            },
            leave: function (node) {
                if (node.kind !== language_1.Kind.FIELD)
                    return;
                cb(node, resultPath, alias);
                resultPath.pop();
                alias.pop();
            }
        });
    }
}
function getLodashDirectiveArgs(node, lodashDirectiveDef) {
    var lodashNode = null;
    for (var _i = 0, _a = node.directives; _i < _a.length; _i++) {
        var directive = _a[_i];
        if (directive.name.value !== lodashDirectiveDef.name)
            continue;
        if (lodashNode)
            throw Error("Duplicating \"@_\" on the \"" + node.name.value + "\"");
        lodashNode = directive;
    }
    if (lodashNode === null)
        return null;
    var args = values_1.getArgumentValues(lodashDirectiveDef, lodashNode);
    return normalizeLodashArgs(lodashNode.arguments, args);
}
function normalizeLodashArgs(argNodes, args) {
    if (!argNodes)
        return args;
    argNodes = keyBy(argNodes, function (argNode) { return argNode.name.value; });
    var orderedArgs = {};
    each(argNodes, function (node, name) {
        var argValue = args[name];
        if (node.value.kind === 'ObjectValue')
            orderedArgs[name] = normalizeLodashArgs(node.value.fields, argValue);
        else if (node.value.kind === 'ListValue') {
            var nodeValues = node.value.values;
            orderedArgs[name] = [];
            for (var i = 0; i < nodeValues.length; ++i)
                orderedArgs[name][i] = normalizeLodashArgs(nodeValues[i].fields, argValue[i]);
        }
        else if (node.value.kind === 'EnumValue' && node.value.value === 'none')
            orderedArgs[name] = undefined;
        else
            orderedArgs[name] = argValue;
    });
    return orderedArgs;
}
function stripQuery(queryAST) {
    return language_1.visit(queryAST, (_a = {},
        _a[language_1.Kind.DIRECTIVE] = function (node) {
            if (node.name.value === '_')
                return null;
        },
        _a));
    var _a;
}
function applyLodashDirective(pathToArgs, data) {
    if (data === null)
        return null;
    var changedData = applyOnPath(data, pathToArgs);
    return applyLodashArgs([], changedData, pathToArgs['@_']);
}
function applyOnPath(result, pathToArgs) {
    var currentPath = [];
    return traverse(result, pathToArgs);
    function traverse(root, pathRoot) {
        if (Array.isArray(root))
            return root.map(function (item) { return traverse(item, pathRoot); });
        if (typeof root === 'object') {
            var changedObject = root ? Object.assign({}, root) : root;
            for (var key in pathRoot) {
                if (key === '@_')
                    continue;
                currentPath.push(key);
                var changedValue = traverse(get(root, key), get(pathRoot, key));
                var lodashArgs = get(pathRoot, [key, '@_'].join('.'));
                changedValue = applyLodashArgs(currentPath, changedValue, lodashArgs);
                if (typeof changedValue === 'function') {
                    changedValue(changedObject, key);
                }
                else {
                    set(changedObject, key, changedValue);
                }
                currentPath.pop();
            }
            return changedObject;
        }
        else {
            return root;
        }
    }
}
function applyLodashArgs(path, object, args) {
    try {
        return transformations_1.applyTransformations(object, args);
    }
    catch (e) {
        console.log(path);
        throw e;
    }
}
//# sourceMappingURL=gql.js.map